EXPORTS
ieee1284_find_ports
ieee1284_free_ports
ieee1284_get_deviceid
ieee1284_open
ieee1284_close
ieee1284_ref
ieee1284_unref
ieee1284_claim
ieee1284_release
ieee1284_get_irq_fd
ieee1284_clear_irq
ieee1284_read_data
ieee1284_write_data
ieee1284_wait_data
ieee1284_data_dir
ieee1284_read_status
ieee1284_wait_status
ieee1284_read_control
ieee1284_write_control
ieee1284_frob_control
ieee1284_do_nack_handshake
ieee1284_negotiate
ieee1284_terminate
ieee1284_ecp_fwd_to_rev
ieee1284_ecp_rev_to_fwd
ieee1284_nibble_read
ieee1284_compat_write
ieee1284_byte_read
ieee1284_epp_read_data
ieee1284_epp_write_data
ieee1284_epp_read_addr
ieee1284_epp_write_addr
ieee1284_ecp_read_data
ieee1284_ecp_write_data
ieee1284_ecp_read_addr
ieee1284_ecp_write_addr
ieee1284_set_timeout
